#include <windows.h>
#include <atlbase.h>
#include <atlconv.h>
#include "MJOGGInputSource.h"

BOOL g_MJSP = FALSE;

/******************************************************************************************
I/O callbacks
******************************************************************************************/
size_t read_func(void *ptr, size_t size, size_t nmemb, void *datasource)
{
	HANDLE file = (HANDLE)datasource;

	IGeneralReader *pReader = (IGeneralReader *) datasource;
	unsigned int bytesread = 0;

	try{
		bytesread = pReader->Read((BYTE *) ptr, size*nmemb);
	}
	catch(...){
		return -1;
	};

	return bytesread/size;
}

int seek_func(void *datasource, ogg_int64_t offset, int whence)
{
	IGeneralReader *pReader = (IGeneralReader *) datasource;
	
	if (g_MJSP)
		return -1;

	// due to limitations of OGG format, fail to seek for remote file.
	if (pReader->GetIsRemote() || pReader->GetIsStream())
		return -1;
	
	HANDLE file = (HANDLE)datasource;
	int seek_type;
	int seek_highword = (int)(offset>>32);

	switch(whence)
	{
		case SEEK_SET:
			seek_type = FILE_BEGIN;
			break;
		case SEEK_CUR:
			seek_type = FILE_CURRENT;
			break;
		case SEEK_END:
			seek_type = FILE_END;
			break;
	}
	
	int nRetVal = pReader->SetPosition((int)(offset&0xffffffff), seek_type);
	if (nRetVal != 0)
		return -1;

	return 0;
}

int close_func(void *datasource)
{
	return 0;
}

long tell_func(void *datasource)
{
	IGeneralReader *pReader = (IGeneralReader *) datasource;
	return pReader->GetPosition();
}


/******************************************************************************************
Exported plugin function
******************************************************************************************/
extern "C"
{
	__declspec (dllexport) CMJInputSource *GetInputSource()
	{
		return new CMJOGGInputSource;
	}
}

/******************************************************************************************
Construction
******************************************************************************************/
CMJOGGInputSource::CMJOGGInputSource() : CMJInputSource()
{
	// intialize variables
	m_pOggInputFile = NULL;
}

/******************************************************************************************
Destruction
******************************************************************************************/
CMJOGGInputSource::~CMJOGGInputSource()
{
	if (m_pOggInputFile)
	{
		ov_clear(m_pOggInputFile);
		delete m_pOggInputFile;
		m_pOggInputFile = NULL;
	}
}

/******************************************************************************************
Open
******************************************************************************************/
int CMJOGGInputSource::Open(LPCTSTR pFilename)
{
	g_MJSP = FALSE;

	if (_tcsstr(pFilename, _T("mjsp://")) != NULL) g_MJSP = TRUE;

	if (CMJInputSource::Open(pFilename) != 0) { return -1; }
	
	// create the vorbis file object
	m_pOggInputFile = new OggVorbis_File;
	if (m_pOggInputFile == NULL) { return -1; }
		
	// set the callbacks
	vorbis_info *vi = NULL;
	ov_callbacks callbacks = {read_func, seek_func, close_func, tell_func};
	if (ov_open_callbacks(m_pReader, m_pOggInputFile, NULL, 0, callbacks) < 0) { return -1; }

	// set the file information
	vi = ov_info(m_pOggInputFile, -1);
	m_nCurrentSampleRate = vi->rate;
	m_nCurrentChannels = vi->channels;
	m_nCurrentBitrate = ov_bitrate(m_pOggInputFile, -1) / 1000;
	m_nCurrentBitsPerSample = 16;

	// error check the format
	if (m_nCurrentChannels > 2) { return -1; }
	
	// set the length
	if (g_MJSP)
		m_nLengthMS = 0;
	else
		m_nLengthMS = (int)ov_time_total(m_pOggInputFile, -1) * 1000;

	return 0;
}

/******************************************************************************************
Add data to the internal buffer
******************************************************************************************/
int CMJOGGInputSource::AddDataToBuffer()
{
	bool bKillFlag = FALSE;
	int nBytesToRead = min(16384, GetBufferBytesAvailable());
	
	int nBytesRead;

	try 
	{
		nBytesRead = ov_read(m_pOggInputFile, (char *) GetBufferFillLocation(), nBytesToRead, 0, 2, 1, &m_nCurrentSection);
		long Bitrate = ov_bitrate_instant(m_pOggInputFile) / 1000;
		if (Bitrate != 0) m_nCurrentBitrate = Bitrate;
	}
	catch(...)
	{
		m_bDone = TRUE;
		return -1;
	}

	UpdateBufferSize(nBytesRead);
	
	if (nBytesRead > 0)
		return 0;
	else
		return -1;
}

/******************************************************************************************
Get information from the plugin
******************************************************************************************/
int CMJOGGInputSource::GetInformation(MJ_INPUT_SOURCE_INFO_FIELDS nInfoID, LPTSTR pResult)
{
	if (nInfoID == MJ_INPUT_SOURCE_INFO_PLUGIN_DESCRIPTION)
	{
		_stprintf(pResult, _T("Media Jukebox OGG Plugin"));
		return 0;
	}

	if (nInfoID == MJ_INPUT_SOURCE_INFO_EXTENSIONS)
	{
		_stprintf(pResult, _T("OGG; Ogg Vorbis File (*.OGG)"));
		return 0;
	}

	return CMJInputSource::GetInformation(nInfoID, pResult);
}	

/******************************************************************************************
Set information
******************************************************************************************/
int CMJOGGInputSource::SetInformation(MJ_INPUT_SOURCE_INFO_FIELDS nInfoID, LPCTSTR pValue)
{
	if (nInfoID == MJ_INPUT_SOURCE_INFO_POSITION_MS)
	{
		// get the position
		int nPositionMS = _ttoi(pValue);
		
		try 
		{
			ov_time_seek(m_pOggInputFile, nPositionMS / 1000);
			ResetInternalBuffer();
		}
		catch(...)
		{
			m_bDone = TRUE;
			return -1;
		}
		
		return 0;
	}

	return CMJInputSource::SetInformation(nInfoID, pValue);
}	

/******************************************************************************************
Perform a command
******************************************************************************************/
int CMJOGGInputSource::DoCommand(MJ_INPUT_SOURCE_COMMANDS nCommandID, int nParam1, int nParam2)
{
	
	return CMJInputSource::DoCommand(nCommandID, nParam1, nParam2);
}