/* VorbisComments.cpp
 **
 ** This program is distributed under the GNU General Public License, version 2.
 ** A copy of this license is included with this source.
 **
 ** Copyright 2001, Warren Spits <spits@cyberdude.com>
 **
 ** A class that holds value tag pairs.
 **/

#include "VorbisComments.h"

#include <codec.h>
#include <string.h> // memcpy, memmove
#include <ctype.h> // toupper

// Initialise the vorbis comments (Constructor)
VorbisComments::VorbisComments() : allocated(8),
                                   count(0)
{
	tags = new tag[allocated];
}

// Destroy the vorbis comments (Destructor)
VorbisComments::~VorbisComments()
{
	for (unsigned int i = 0; i < count; i++)
	{
		delete tags[i].name;
		delete tags[i].value;
	}
	delete tags;
}

// Get a tags name
char * VorbisComments::getName(unsigned int item) const
{
	char *result = 0;

	if (item < count)
		result = tags[item].name;

	return result;
}

// Get a tags value
char * VorbisComments::getValue(unsigned int item) const
{
	char *result = 0;

	if (item < count)
		result = tags[item].value;

	return result;
}

// Get a tags value
char * VorbisComments::getValue(char * name) const
{
	char *result = 0;
	
	for (unsigned int i = 0; i < count; i++)
	{
		if(strcmp(tags[i].name, name) == 0)
		{
			result = getValue(i);
			break;
		}
	}
	
	return result;
}


// Returns the number of tags
unsigned int VorbisComments::getCount() const
{
	return count;
}

// Adds a tag
void VorbisComments::addTag(const char *name, const char *value)
{
	size_t len;
	char *upper;

	if (count == allocated)
	{
		// Increase the allocated size
		tag *newtags;

		newtags = new tag[allocated*2];
		memcpy(newtags, tags, sizeof(tag) * allocated);
		allocated *= 2;
		delete tags;
		tags = newtags;
	}

	// Add the name
	len = strlen(name);
	tags[count].name = new char[len +1];
	strcpy(tags[count].name, name);
	// Upper case the name
	upper = tags[count].name;
	while (*upper != 0)
		*upper++ = toupper(*upper);

	// Add the value
	len = strlen(value);
	tags[count].value = new char[len + 1];
	strcpy(tags[count].value, value);

	count++;
}

// Removes a tag
void VorbisComments::removeTag(unsigned int item)
{
	if (item < count)
	{
		delete tags[item].name;
		delete tags[item].value;
		memmove(tags + item, tags + item + 1, sizeof(tag) * (count - item - 1));
		count--;
	}
}

// Get a tags value
void VorbisComments::removeTag(const char * name)
{
	bool bAgain = true;
	while (bAgain)
	{
		bAgain = false;
		
		for (unsigned int i = 0; i < count; i++)
		{
			if(strcmp(tags[i].name, name) == 0)
			{
				removeTag(i);
				bAgain = true;
				break;
			}
		}
	}
}

// Remove all the tags
void VorbisComments::removeAll()
{
	for (unsigned int i = 0; i < count; i++)
	{
		delete tags[i].name;
		delete tags[i].value;
	}
	count = 0;
}

// Retrieve the comments from a vorbis_comment structure
void VorbisComments::retrieve(const vorbis_comment *vc)
{
	char *comment;
	char *equalpos;

	removeAll();

	for (int i = 0; i < vc->comments; i++)
	{
		// Copy the comment
		comment = new char[vc->comment_lengths[i] +1];
		strcpy(comment, vc->user_comments[i]);

		// Find the '=' and set it to NULL
		equalpos = strchr(comment, '=');
		if(equalpos)
		{
			*equalpos = 0;

			// Add the tag
			addTag(comment, equalpos +1); 
		}
		
		delete comment;
	}

}

// Store the comments in a vorbis_comment structure
void VorbisComments::store(vorbis_comment *vc) const
{
	vorbis_comment_clear(vc);
	vorbis_comment_init(vc);

	for (unsigned int i = 0; i < count; i++)
		vorbis_comment_add_tag(vc, tags[i].name, tags[i].value);
}

// Swap two comments around
void VorbisComments::swapPos(unsigned int pos1, unsigned int pos2)
{
	tag temp;

	if ((pos1 != pos2) && (pos1 < count) && (pos2 < count))
	{
		temp = tags[pos1];
		tags[pos1] = tags[pos2];
		tags[pos2] = temp;
	}
}

// Modify a comment
void VorbisComments::modifyTag(unsigned int item, const char *name, const char *value)
{
	size_t len;

	if (item < count)
	{
		// Set the name
		delete tags[item].name;
		len = strlen(name);
		tags[item].name = new char[len +1];
		strcpy(tags[item].name, name);

		// Set the value
		delete tags[item].value;
		len = strlen(value);
		tags[item].value = new char[len +1];
		strcpy(tags[item].value, value);
	}
}

// Checks a name for validity
bool VorbisComments::isValidName(const char *name)
{
	const char *pos;

	if (strlen(name) == 0)
		return false;

	pos = name;
	while (*pos != '\0')
	{
		if ((*pos < 0x20) || (*pos > 0x7D) || (*pos == '='))
			return false;
		pos++;
	}

	return true;	
}

// Checks a value for validity
bool VorbisComments::isValidValue(const char *value)
{
	if (strlen(value) == 0)
		return false;

	return true;
}
